=begin

Copyright 2011-2014, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         MoleculeImporter.rb
Author:       Andreas Eisenbarth
Usage:        menu Plugins > 'Molecule Importer'
Description:  This Plugin allows to import chemical structures in *.mol or
              z-matrix file format. The resulting structure from z-matrices
              will not have chemical bonds, since they are not existing in t
              hat format.
Version:      1.2.0
Date:         05.12.2014

=end
require 'sketchup.rb'


module AE


module MoleculeImporter


unless defined?(self::PLUGIN_ROOT)

  PLUGIN_ROOT = File.dirname(__FILE__)
  # Translation
  require File.join(PLUGIN_ROOT, "Translate.rb")
  TRANSLATE = Translate.new("MoleculeImporter", File.join(PLUGIN_ROOT, "lang"))

  # Chemical parameters
  PERIODIC_SYSTEM = ["H", "He", "Li", "Be", "B", "C", "N", "O", "F", "Ne", "Na", "Mg", "Al", "Si", "P", "S", "Cl", "Ar", "K", "Ca", "Sc", "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb", "Sr", "Y", "Zr", "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te", "I", "Xe", "Cs", "Ba", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Hf", "Ta", "W", "Re", "Os", "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th", "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr", "Rf", "Db", "Sg", "Bh", "Hs", "Mt", "Ds", "Rg", "Cn", "Uut", "Uuq", "Uup", "Uuh", "Uus", "Uuo"]
  # Average covalent atomic radii in picometer; source: http://en.wikipedia.org/wiki/Atomic_radius (empiric, or calculated)
  # Note: Accurate atom radii depend very much on charge, hybridization and binding partners!
  ATOMIC_RADIUS = {"H"=>25, "He"=>31, "Li"=>145, "Be"=>105, "B"=>85, "C"=>70, "N"=>65, "O"=>60, "F"=>50, "Ne"=>38, "Na"=>180, "Mg"=>150, "Al"=>125, "Si"=>110, "P"=>100, "S"=>100, "Cl"=>100, "Ar"=>71, "K"=>220, "Ca"=>180, "Sc"=>160, "Ti"=>140, "V"=>135, "Cr"=>140, "Mn"=>140, "Fe"=>140, "Co"=>135, "Ni"=>135, "Cu"=>135, "Zn"=>135, "Ga"=>130, "Ge"=>125, "As"=>115, "Se"=>115, "Br"=>115, "Kr"=>88, "Rb"=>235, "Sr"=>200, "Y"=>180, "Zr"=>155, "Nb"=>145, "Mo"=>145, "Tc"=>135, "Ru"=>130, "Rh"=>135, "Pd"=>140, "Ag"=>160, "Cd"=>155, "In"=>155, "Sn"=>145, "Sb"=>145, "Te"=>140, "I"=>140, "Xe"=>108, "Cs"=>260, "Ba"=>215, "La"=>195, "Ce"=>185, "Pr"=>185, "Nd"=>185, "Pm"=>185, "Sm"=>185, "Eu"=>185, "Gd"=>180, "Tb"=>175, "Dy"=>175, "Ho"=>175, "Er"=>175, "Tm"=>175, "Yb"=>175, "Lu"=>175, "Hf"=>155, "Ta"=>145, "W"=>135, "Re"=>135, "Os"=>130, "Ir"=>135, "Pt"=>135, "Au"=>135, "Hg"=>150, "Tl"=>190, "Pb"=>180, "Bi"=>160, "Po"=>190, "At"=>130, "Rn"=>120, "Fr"=>260, "Ra"=>215, "Ac"=>195, "Th"=>180, "Pa"=>175, "U"=>175, "Np"=>175, "Pu"=>175, "Am"=>175, "Cm"=>175, "Bk"=>175, "Cf"=>175, "Es"=>175, "Fm"=>175, "Md"=>175, "No"=>175, "Lr"=>175, "Rf"=>175, "Db"=>175, "Sg"=>175, "Bh"=>175, "Hs"=>175, "Mt"=>175, "Ds"=>175, "Rg"=>175, "Cn"=>175, "Uut"=>220, "Uuq"=>220, "Uup"=>220, "Uuh"=>220, "Uus"=>220, "Uuo"=>220}
  COLORS = {"H"=>"dddddd", "He"=>"ff8888", "Li"=>"99bbcc", "Be"=>"8899cc", "B"=>"2244aa", "C"=>"222222", "N"=>"dd0000", "O"=>"0000dd", "F"=>"00cc55", "Ne"=>"ddbb44", "Na"=>"aa99ff", "Mg"=>"cc66cc", "P"=>"6600aa", "S"=>"33ffff", "Cl"=>"44cc33", "Cr"=>"ee3355", "Mn"=>"ee0020", "Fe"=>"cc3300", "Co"=>"0047AB", "Ni"=>"00aab0", "Cu"=>"e07033", "Br"=>"99cc00", "Pd"=>"000000", "Ag"=>"aaaaaa", "I"=>"990055", "Pt"=>"000000", "Au"=>"eebb00", "Hg"=>"888888"}

  # Drawing parameters
  COMPONENT_ATOM = File.join(PLUGIN_ROOT, "skp", "atom_high.skp")
  COMPONENT_SINGLE_BOND = File.join(PLUGIN_ROOT, "skp", "bond1.skp")
  COMPONENT_DOUBLE_BOND = File.join(PLUGIN_ROOT, "skp", "bond2.skp")
  COMPONENT_TRIPLE_BOND = File.join(PLUGIN_ROOT, "skp", "bond3.skp")
  # TODO: How to handle bond length of mixed bonds like aromatic bonds, rings, delocalization?
  # General scale for complete molecule. 
  # 1 Ångström ≈ x * 1 m, but most files don't contain real-world dimensions.
  GLOBAL_SCALE = 1.0
  # Tweak this to tune between skeletal model (0), ball-and-stick model (1), space-filling model (big number)
  ATOM_SCALE = 1.0
  # Radius of bond sticks, fraction of ATOM_SCALE.
  BOND_SCALE = 0.5

end


# Inserts components for atoms and bonds
# @param [Hash] hash of atoms and bonds
#
def self.draw_molecule(hash)
  model = Sketchup.active_model
  entities = model.entities
  if Sketchup.version_number > 7000000
    model.start_operation(TRANSLATE["Import Molecule"], true) # SU7.0+
  else
    model.start_operation(TRANSLATE["Import Molecule"])
  end

  # Load components or find them if already loaded
  atom_definition = model.definitions[File.basename(COMPONENT_ATOM)] || model.definitions.load(COMPONENT_ATOM)
  bond_definition = {
    1 => model.definitions[File.basename(COMPONENT_SINGLE_BOND)] || model.definitions.load(COMPONENT_SINGLE_BOND),
    2 => model.definitions[File.basename(COMPONENT_DOUBLE_BOND)] || model.definitions.load(COMPONENT_DOUBLE_BOND),
    3 => model.definitions[File.basename(COMPONENT_TRIPLE_BOND)] || model.definitions.load(COMPONENT_TRIPLE_BOND)
  }

  # Inspect the atom distance to check if the source file format uses picometer or random units
  unless hash[:atoms].empty? || hash[:bonds].empty?
    first_bond = hash[:bonds][1] # key 1 of a hash
    atom1 = first_bond[:atom1]
    atom2 = first_bond[:atom2]
    first_length = (hash[:atoms][atom1][:coords]).distance(hash[:atoms][atom2][:coords])
    # If source file format uses Ångström, convert to picometer
    if (0.95..3).include?(first_length)
      scale = 100 * GLOBAL_SCALE
    # If source file format uses random units, set a fixed scale
    elsif !(95..300).include?(first_length)
      scale = 150/first_length * GLOBAL_SCALE
    end
  end

  # Loop over all atoms, insert atom components, paint them according to element
  if ATOM_SCALE != 0
    group = entities.add_group
    group.name = hash[:name]
    hash[:atoms].each_pair{ |id, data|
      # look up atom radius of element
      # accurate atom radii depend very much on charge, hybridization and binding partners!
      element = data[:element]
      radius = ATOMIC_RADIUS[element] || 150
      # SketchUp Bug: we need to apply the scaling to each axis and not one overall scale factor.
      coords = data[:coords].transform( Geom::Transformation.scaling(*[scale / 2.54]*3) )
      # SketchUp Bug: we need to apply the scaling to each axis and not one overall scale factor.
      scaling = Geom::Transformation.scaling(*[radius * ATOM_SCALE]*3) 
      translation = Geom::Transformation.translation(coords)
      transform = translation * scaling
      atom = group.entities.add_instance(atom_definition, transform)
      atom.transform!(Geom::Transformation.new(coords))
      atom.material = eval("0x00"+COLORS[element]) if COLORS[element]
      atom.name = element+id.to_s
      atom.set_attribute("Chemical", "id", id)
      atom.set_attribute("Chemical", "element", element)
    } # each_pair
  end # if

  # Loop over all bonds, insert bond components.
  hash[:bonds].each_pair{ |id, data|
    atom1 = data[:atom1]
    atom2 = data[:atom2]
    coord1 = hash[:atoms][atom1][:coords].collect{ |x| x.cm * scale }
    coord2 = hash[:atoms][atom2][:coords].collect{ |x| x.cm * scale }
    length = coord1.distance(coord2)
    vector = coord1.vector_to coord2
    radius = ATOM_SCALE * BOND_SCALE
    # 3 vectors needed for axes transformation to set proper orientation of bond along vector
    tmp_normal_vector = Geom::Vector3d.new(vector.y,vector.z,vector.x) # not yet orthogonal
    # rotate double bond correctly
    if data[:order] == 2
      neighbour_coords = []
      hash[:bonds].each_value{ |d|
        if d[:atom1] == (atom1||atom2)
          neighbour_atom = d[:atom2]
          neighbour_coords.push(hash[:atoms][neighbour_atom][:coords])
        elsif d[:atom2] == (atom1||atom2)
          neighbour_atom = d[:atom1]
          neighbour_coords.push(hash[:atoms][neighbour_atom][:coords])
        end
      }
      # Here: normal vector (pi orbitals) should stand orthogonal on double bond.
      tmp_normal_vector = Geom::Vector3d.new(Geom.fit_plane_to_points(neighbour_coords).slice(0,3)) if neighbour_coords.length >= 3
    end # if double bond
    normal_vector2 = vector * tmp_normal_vector
    normal_vector1 = normal_vector2 * vector
    axes = Geom::Transformation.axes(coord1, normal_vector1, normal_vector2, vector)
    translation = Geom::Transformation.translation(coord1)
    scaling = Geom::Transformation.scaling(radius*scale, radius*scale, 2*2.54*length)
    transform = translation * axes * scaling
    definition = bond_definition[data[:order]] || bond_definition[1] # single, double, triple bond
    bond = group.entities.add_instance(definition, transform)
    bond.set_attribute("Chemical", "id", id)
    bond.set_attribute("Chemical", "length", length)
    bond.set_attribute("Chemical", "atom1", atom1)
    bond.set_attribute("Chemical", "atom2", atom2)
  } # each_pair

  min = group.local_bounds.min.z
  group.transform!( Geom::Transformation.new([0, 0, -min]) )
  model.layers["handles"].visible = false
  model.commit_operation
end # def draw_molecule



class MolImporter < Sketchup::Importer

  IMPORT_SUCCESS        = 0
  IMPORT_FAILED         = 1
  IMPORT_FILE_NOT_FOUND = 4

  def description
    return "MDL Molfile [v3000, v2000] (*.mol)"
  end

  def file_extension
    return "mol"
  end

  def id
    return "com.sketchup.importers.mol_file"
  end

  def supports_options?
    return false
  end

  def load_file(path, status=true)
    return IMPORT_FILE_NOT_FOUND unless path.is_a?(String)
    path = File.expand_path(path)
    return IMPORT_FILE_NOT_FOUND unless File.file?(path)
    @filename = File.basename(path, ".mol")
    hash = parse_mol(IO.read(path))
    MoleculeImporter.draw_molecule(hash)
    Sketchup.send_action("viewZoomExtents:")
    return IMPORT_SUCCESS
  rescue Exception => e
    puts(e.message)
    puts(e.backtrace.join("\n"))
    return IMPORT_FAILED
  end

  # Reads MDL Molfiles, two versions.
  # Creates a hash of all atoms and all bonds to draw.
  # @param [String] a string in mol format (the file's contents)
  #
  def parse_mol(filestring)
    lines = filestring.split($/)
    # header
    hash = {:name => @filename, :atoms => {}, :bonds => {}}
    # Split general info line at max 3 symbols except last value that can be longer
    # info = lines[3].to_s.scan(/...(?!\w*$)|\w+$/).collect{|x| x.strip}
    info = lines[3].to_s.strip.split(/\s+/)
    lines.slice!(0,4)
    version = info.last

    # Older version V2000
    if version == "V2000"
      atoms_count = info[0].to_i
      bonds_count = info[1].to_i
      for i in (1..atoms_count)
        data = lines.shift.to_s
        coords = [data[0,10].to_f, data[10,10].to_f, data[20,10].to_f]
        hash[:atoms][i] = {:element => data[30,3].strip, :coords => coords}
      end
      for i in (1..bonds_count)
        # split line at each 3 places
        data = lines.shift.to_s.scan(/.../)
        hash[:bonds][i] = {:atom1 => data[0].to_i, :atom2 => data[1].to_i, :order => data[2].to_i}
      end

    # New version V3000
    elsif version == "V3000"
      data = lines.shift.to_s.strip.split(/\s+/)  until data.to_a.length > 5 && data[2] == "COUNTS"
      atoms_count = data[3].to_i
      bonds_count = data[4].to_i
      # Find ATOM section
      data = ""
      data = lines.shift while data[/BEGIN/].nil? && data[/ATOM/].nil?
      for i in (1..atoms_count)
        data = lines.shift
        next if !data.sub!(/^M\s\sV30\s/,"") # skip comments that don't start with M  V30
        data = data.sub(/–$|-$/,"") + lines.shift.sub(/^M\s\sV30\s+/,"") until !data[/–$|-$/] # connect broken lines (wider than 80 characters)
        data = data.strip.split(/\s+/)
        coords = data.slice(2,3).collect!{|x| x.to_f}
        hash[:atoms][data[0].to_i] = {:element => data[1], :coords => coords}
      end
      # Find BOND section
      data = ""
      data = lines.shift while data[/BEGIN/].nil? && data[/BOND/].nil?
      for i in (1..bonds_count)
        data = lines.shift
        next if !data.sub!(/^M\s\sV30\s/,"")
        data = data.sub(/–$|-$/,"") + lines.shift.sub(/^M\s\sV30\s+/,"") until !data[/–$|-$/] # connect broken lines (wider than 80 characters)
        data = data.strip.split(/\s+/)
        hash[:bonds][data[0].to_i] = {:atom1 => data[2].to_i, :atom2 => data[3].to_i, :order => data[1].to_i}
      end
    end
    return hash
  end # def parse_mol
  private :parse_mol

end
Sketchup.register_importer(MolImporter.new)



class ZImporter < Sketchup::Importer

  IMPORT_SUCCESS        = 0
  IMPORT_FAILED         = 1
  IMPORT_FILE_NOT_FOUND = 4

  def description
    return "Z-Matrix (GAUSSIAN, MOPAC, *.mop, *.zmt, *.dat, *.mpc)"
  end

  def file_extension
    return "*"
  end

  def id
    return "com.sketchup.importers.z_matrix"
  end

  def supports_options?
    return false
  end

  def load_file(path, status=true)
    return IMPORT_FILE_NOT_FOUND unless path.is_a?(String)
    file = File.expand_path(path)
    return IMPORT_FILE_NOT_FOUND unless File.file?(path)
    @filename = File.basename(path)
    hash = parse_z(IO.read(path))
    MoleculeImporter.draw_molecule(hash)
    Sketchup.send_action("viewZoomExtents:")
    return IMPORT_SUCCESS
  rescue Exception => e
    $stderr.write(e.message)
    $stderr.write(e.backtrace.join("\n"))
    return IMPORT_FAILED
  end

  # Reads Z-matrix, Gaussian and Mopac convention.
  # Converts internal coordinates to cartesic coordinates
  # and creates a hash of all atoms and all bonds to draw.
  # @param [String] a string in z-matrix format (the file's contents)
  #
  def parse_z(filestring)
    hash = {:name => @filename, :atoms => {}, :bonds => {}}
    # skip until data section (GAMESS)
    # and remove first 3 uncommented lines that contain no data
    filestring.sub!(/^(.|\n)*\$DATA(.*\n){2}|^(.*\n){3}/, "")
    datasections = filestring.split(/\n\s*\n/) | [filestring]

    # some z-matrices use named variables instead of numbers, and a variable list at the end after an empty line
    # because of that, read the variables first
    if datasections.length > 1
      variable_list = {}
      lines = datasections[1].split($/)
      # remove comments
      lines.reject!{|l| l[0,1]==("*"||"#") }
      lines.each{|l| a=l.split(/\s+/); variable_list[a[0].strip]=a[1].to_f}
    end

    # main section / z-matrix
    lines = datasections[0].split($/)
    # remove comments
    lines.reject!{|l| l[0,1][/[^a-zA-Z0-9]/] }
    # detect type of z-matrix
    if lines[3].strip.split(/\s+/).length==7# && !type
      type = :gaussian
    elsif lines[3].strip.split(/\s+/).length==10# && !type
      type = :mopac
    end

    n = 1 # atom index number
    for i in (1..lines.length)
      line = lines.shift
      data = line.to_s.strip.split(/\s+/)

      # sort data
      if type == :gaussian
        # Gaussian order:  element(+number?), ref, bond length, ref, bond angle, ref, dihedral angle
        # => ref, bond length, ref, bond angle, ref, dihedral angle
        element = data[0][/[A-Za-z]+/]
        data.shift
        c=[]
        data.each_with_index{ |x, j|
          # detect named variables and assign numbers
          if x[/[a-zA-Z]/]
            c[j] = variable_list[x]
          elsif (j%2) == 0
            c[j] = x.to_i
          elsif (j%2) == 1
            c[j] = x.to_f
          end
        }
      elsif type == :mopac
        # MOPAC order:  element, 1, bond length, 1, bond angle, 1, dihedral angle, reference atom for bond length, ref for bond angle, ref for dihedral angle
        # => ref, bond length, ref, bond angle, ref, dihedral angle
        element = data[0]
        c = [data[7].to_i, data[1].to_f, data[8].to_i, data[3].to_f, data[9].to_i, data[5].to_f]
      end

      # first atom: no coordinates specified => set to origin
      if n == 1
        coords = [0,0,0]
      elsif n == 2 # second atom: bond length to first atom => set along x direction
        length = 100 * c[1] # bond length; convert Ångström to pm
        vector = Geom::Vector3d.new(length, 0, 0)
        coords = Geom::Point3d.new(hash[:atoms][c[0]][:coords]) + vector

      # third atom: bond length and bond angle => set in xy plane
      elsif n == 3
        length = 100 * c[1] # bond length; convert Ångström to pm
        alpha = c[3].degrees # bond angle
        vector = Geom::Vector3d.new(-Math.cos(alpha), Math.sin(alpha), 0)
        vector.length = length
        coords = Geom::Point3d.new(hash[:atoms][c[0]][:coords]) + vector

      # all other atoms: complete set of internal coordinates
      else
        length = 100 * c[1] # bond length; convert Ångström to pm
        alpha = c[3].degrees # bond angle
        delta = c[5].degrees # dihedral angle
        x = -Math.cos(alpha)
        y = Math.sin(alpha) * Math.cos(delta)
        z = Math.sin(alpha) * Math.sin(delta)
        # vector towards next atom, in local coordinate system (x-axis, xy-plane)
        vector = Geom::Vector3d.new(x, y, z)
        # transform/rotate vector to fit into global coordinate system (current bond, plane of two bonds)
        b1 = hash[:atoms][c[2]][:coords].vector_to hash[:atoms][c[0]][:coords]
        b2 = hash[:atoms][c[2]][:coords].vector_to hash[:atoms][c[4]][:coords]
        b3 = b1 * b2
        b2 = b3 * b1
        t = Geom::Transformation.axes [0,0,0], b1, b2, b3
        vector.transform!(t)
        vector.length = length
        coords = Geom::Point3d.new(hash[:atoms][c[0]][:coords]) + vector
      end

      hash[:atoms][n] = {:element => element, :coords => coords}
      n += 1
    end
    return hash
  end # def parse_z
  private :parse_z

end
Sketchup.register_importer(ZImporter.new)


unless file_loaded?(__FILE__)

  # Command for Molfile
  cmd_import_mol = UI::Command.new(TRANSLATE["Import MDL Molfile"] + " [v3000, v2000] (*.mol)"){
    openpanel_title = TRANSLATE["Open MDL Molfile"]
    path = UI.openpanel(openpanel_title, ".mol")
    success = MolImporter.new.load_file(path)
    case success
    when MolImporter::IMPORT_FILE_NOT_FOUND
      UI.messagebox(TRANSLATE["File seems not to exist.\n Please check that the file's location does not contain Unicode characters. This is a limitation of SketchUp and its Ruby version."], MB_OK)
    when MolImporter::IMPORT_FAILED
      UI.messagebox(TRANSLATE["Import failed."], MB_OK)
    end
  }
  cmd_import_mol.tooltip = TRANSLATE["Imports chemical molecule models from the MDL Molfile format."]
  cmd_import_mol.small_icon = File.join(PLUGIN_ROOT, "images", "icon_moleculeimporter_16.png")
  cmd_import_mol.large_icon = File.join(PLUGIN_ROOT, "images", "icon_moleculeimporter_24.png")


  # Command for Z matrix
  cmd_import_z = UI::Command.new(TRANSLATE["Import Z-Matrix"] + " (GAUSSIAN, MOPAC, *.mop, *.zmt, *.dat, *.mpc)") {
    openpanel_title = TRANSLATE["Open Z matrix file"]
    path = UI.openpanel(openpanel_title, ".*")
    success = ZImporter.new.load_file(path)
    case success
    when ZImporter::IMPORT_FILE_NOT_FOUND
      UI.messagebox(TRANSLATE["File seems not to exist.\n Please check that the file's location does not contain Unicode characters. This is a limitation of SketchUp and its Ruby version."], MB_OK)
    when ZImporter::IMPORT_FAILED
      UI.messagebox(TRANSLATE["Import failed."], MB_OK)
    end
  }
  cmd_import_z.tooltip = TRANSLATE["Imports chemical molecule models from the Z-Matrix format."]
  cmd_import_z.small_icon = File.join(PLUGIN_ROOT, "images", "icon_moleculeimporter_16.png")
  cmd_import_z.large_icon = File.join(PLUGIN_ROOT, "images", "icon_moleculeimporter_24.png")


  mol_importer_menu = UI.menu("Plugins").add_submenu(TRANSLATE["Molecule Importer"])
  mol_importer_menu.add_item(cmd_import_mol)
  mol_importer_menu.add_item(cmd_import_z)

  file_loaded(__FILE__)
end


end # module MoleculeImporter


end # module AE
